/*
	KUMAamp project()
	Copyright (C) Mamiya 2000.
	License:LGPL
	---------------------------------------------------------------------------
	GYM/GEZ sequencer
	seq_gez.cpp
*/

#include "in_gez.h"

/* ANSI/Windows standard headers */
#include <windows.h>
#include <shlobj.h>
#include <stdio.h>

/* Libraries headers */
#if USE_ZLIB
#include <zlib.h>
extern "C" void GYI_load(char *lpszSrcFile, int *ploop, char *title);
extern "C" void *GEZ_extract(LPCSTR lpszSrcFile, int *psize, int *ploop, char *title);
#endif

#include "../m88/opna.h"
#include "../mame/sn76496.h"

/* Project headers */
#include "../sequencer.h"
#include "in_gez.rh"

#define MAX_CH 16
#define OPNA_CLOCK (7670442)	/* 68000̓̓NbN */
#define SN_CLOCK (3579530)		/* Z80̓̓NbN */
#define SOUND_55K (55555)
#define SOUND_48K (48000)
#define SOUND_44K (44100)
#define SOUND_24K (24000)
#define SOUND_22K (22050)
#define SOUND_11K (11025)
#define FADE_MAX (64)
#define FADE_SHIFT (20)

/* GEZ file header struct */
#define GYM_MAGIC (0x584d5947)	/* 'GYMX' */
#define GYM_OFS_MAGIC		(0)
#define GYM_OFS_SONGTITLE	(GYM_OFS_MAGIC +      4)
#define GYM_OFS_GAMETITLE	(GYM_OFS_SONGTITLE + 32)
#define GYM_OFS_PUBLISHER	(GYM_OFS_GAMETITLE + 32)
#define GYM_OFS_GENERATOR	(GYM_OFS_PUBLISHER + 32)
#define GYM_OFS_DUMPER		(GYM_OFS_GENERATOR + 32)
#define GYM_OFS_COMMENTS	(GYM_OFS_DUMPER +    32)
#define GYM_OFS_LOOPTIME	(GYM_OFS_COMMENTS + 256)
#define GYM_OFS_COMPSIZE	(GYM_OFS_LOOPTIME +   4)
#define GYM_OFS_DATASTART	(GYM_OFS_COMPSIZE +   4)


/* GEZ sequencer struct */
typedef struct {
	SEQUENCER seq;			/* KUMAamp common interface */
	int loopc;				/* zero based loopcounter */
	int loopcount;			/* max loop count */
	int frequency;			/* mixing frequency */
	int fade;				/* 0:normal 1-:fading out(dB/(1 >> FADE_SHIFT)) */
	int fadetime;			/* fadeout time(ms) */
	int voldown;			/* (dB) */

	int volFM;				/* (dB) */
	int volSSG;				/* (dB) */
	int volPCM;				/* (dB) */
	/*int volRHYTHM;*/		/* (dB) */

	char title[MAX_PATH];	/* songname */
	unsigned char *data;	/* data pointer top of file */
	unsigned char *ptr;		/* data pointer current */

	/* loop */
	unsigned char *loopptr;	/* data pointer loop */
	int loopclock;			/* loop point (1/60sec) */

#if EXLOOP
	unsigned int saveflagm[8];
	unsigned char savem[256];
	unsigned int saveflags[8];
	unsigned char saves[256];
	unsigned int savekey[8];
#endif

	int clock;				/* played time(1/60sec) */
	int spd1;				/* speedH(samples per 1/ 60sec) */
	int spd2;				/* speedL(samples per 1/360sec) */
	int left1;				/* stepH (samples per 1/ 60sec) */
	int left2;				/* stepL (samples per 1/360sec) */
	int left3;				/* stepH fmsync */
	int frequency2;			/* (real) output frequency(Hz) */
	int fmtimersync;		/* 1:OPNA-timer sync mode 0:compatible mode */

	int fadespd;			/* fadeout speed(dB/(1 >> FADE_SHIFT) per 1/60sec) */
	int isplaying;			/* 1:playing 0:stoped */
} GEZSEQ;

/* GEZamp setting struct */
static struct {
	/* OPNA rythm wav files path */
#if 0
	char rhythmpath[MAX_PATH];
#endif
	char currentrhythmpath[MAX_PATH];
	int frequency;			/* mixing frequency */
	int priority;			/* play thread priority */
	int loopcount;			/* max loop count */
	int fadetime;			/* fadeout time(ms) */
	int fmtimersync;		/* 1:OPNA-timer sync mode 0:compatible mode */

	int volFM;				/* (dB) */
	int volSSG;				/* (dB) */
	int volPCM	;			/* (dB) */
	/*int volRHYTHM;*/		/* (dB) */

	int chenable[MAX_CH];

	HINSTANCE hDllInstance;
	char cfgname[MAX_PATH];
} setting;

typedef struct {
	char *key;
	int num;
} KEYINT;

static FM::OPN2 opn2;

static KEYINT freqlist[] = {
	{ "55467:FM",	SOUND_55K },
	{ "48000",		SOUND_48K },
	{ "44100",		SOUND_44K },
	{ "24000",		SOUND_24K },
	{ "22050",		SOUND_22K },
	{ "11025",		SOUND_11K },
	{ NULL,			SOUND_55K },
};

static KEYINT priolist[] = {
	{ "Highest",	5 },
	{ "Higher",		4 },
	{ "Normal",		3 },
	{ "Lower",		2 },
	{ "Lowest",		1 },
	{ NULL,			3 },
};

#define PTR__ ((GEZSEQ *)(this__->work))
static void Play(SEQUENCER *this__)
{
	PTR__->volFM = PTR__->volSSG = PTR__->volPCM = -9999;
#if 0
	PTR__->volRHYTHM = -9999;

	if (lstrcmpi(setting.currentrhythmpath, setting.rhythmpath))
	{
		OutputDebugString(setting.rhythmpath);
		opn2.LoadRhythmSample(setting.rhythmpath);
		lstrcpy(setting.currentrhythmpath, setting.rhythmpath);
	}
#endif
	PTR__->loopc = 0;
	PTR__->loopcount = setting.loopcount;
	PTR__->fadetime = setting.fadetime;
	PTR__->fade = 0;
	opn2.SetRate(OPNA_CLOCK / 2, PTR__->frequency2, (PTR__->frequency > SOUND_44K));
    SN76496_init(0, SN_CLOCK, PTR__->frequency2);

	PTR__->voldown = 0;
	PTR__->fmtimersync = setting.fmtimersync;
	PTR__->clock = 0;
	PTR__->left1 = 0;
	PTR__->left2 = 0;
	PTR__->left3 = 0;
	PTR__->spd1 = (PTR__->frequency2) / 60;			/* VSYNC */
	PTR__->spd2 = (PTR__->frequency2) % 60;
	PTR__->ptr = PTR__->data;
	PTR__->isplaying = 1;
	opn2.Reset();
	opn2.SetReg(0x29, 0x9f);
	PTR__->loopptr = NULL;
#if EXLOOP
	{
		int i;
		for (i = 0; i < 8; i++)
		{
			PTR__->saveflagm[i] = 0;
			PTR__->saveflags[i] = 0;
			PTR__->savekey[i] = 0x100;
		}
	}
#endif
}
static void Stop(SEQUENCER *this__)
{
	PTR__->loopc = 0;
	PTR__->clock = 0;
	PTR__->left1 = 0;
	PTR__->left2 = 0;
	PTR__->ptr = NULL;
	PTR__->isplaying = 0;
}

#define SETTING_XY(x,y) \
	if (PTR__->y != setting.y) \
	{ \
		PTR__->y = setting.y; \
		opn2.x(- PTR__->voldown - PTR__->y); \
	}
#define SETTING_XYA(x,y) \
	if (1) \
	{ \
		PTR__->y = setting.y; \
		opn2.x(- PTR__->voldown - PTR__->y); \
	}
#define MIXBUF_SIZE SOUND_48K

static void Mix(SEQUENCER *this__, void *buf, int numsamp)
{
	static int mixbuf[MIXBUF_SIZE*8];
	static short snbuf[MIXBUF_SIZE*2];
	short *bufp = (short *)buf;
	unsigned char *p = PTR__->ptr;
	if (p == NULL)
	{
		if (bufp != NULL) ZeroMemory(bufp, numsamp*4);
		return;
	}

	if (bufp != NULL)
	{
		SETTING_XY(SetVolumeFM,       volFM);
		SETTING_XY(SetVolumePSG,      volSSG);
		SETTING_XY(SetVolumePCM,      volPCM);
#if 0
		SETTING_XY(SetVolumeRhythmTotal,volRHYTHM);
#endif
	}

	while (bufp == NULL || numsamp)
	{
		int size;
		while (PTR__->left1 == 0) {
			switch (*p)
			{
				case 0x01:	/* OPNA(m) */
					if (PTR__->fmtimersync && p[1] == 0x27 && (p[2] & 0x30))
					{
						/* sync with fmtimer-reset */
						PTR__->left1 = PTR__->left3;
						PTR__->left3 = 0;
					}
#if EXLOOP
					if (PTR__->loopclock && PTR__->loopptr == NULL)
					{
						PTR__->saveflagm[p[1] >> 5] |= 1 << (p[1] & 31);
						PTR__->savem[p[1]] = p[2];
					}
					if (p[1] == 0x28)
						PTR__->savekey[p[2] & 7] = p[2];
#endif
					opn2.SetReg(p[1], p[2]);
					p += 3;
					break;
				case 0x02:	/* OPNA(s) */
#if EXLOOP
					if (PTR__->loopclock && PTR__->loopptr == NULL)
					{
						PTR__->saveflags[p[1] >> 5] |= 1 << (p[1] & 31);
						PTR__->saves[p[1]] = p[2];
					}
#endif
					opn2.SetReg(0x100 | p[1], p[2]);
					p += 3;
					break;
				case 0x03:	/* PSG */
					SN76496Write(0, p[1]);
					p += 2;
					break;
				case 0x00:	/* SYNC(1) */
					if (PTR__->fmtimersync)
					{
						PTR__->left1 = PTR__->left3;
						PTR__->left3 = PTR__->spd1;
					}
					else
					{
						PTR__->left1 = PTR__->spd1;
					}
					PTR__->left2 += PTR__->spd2;
					PTR__->clock += 1;
					p++;
					if (PTR__->loopclock && PTR__->loopptr == NULL && PTR__->clock == PTR__->loopclock)
					{
						PTR__->loopptr = p;
					}
					if (PTR__->fade)
					{
						if (PTR__->fade > (FADE_MAX << FADE_SHIFT))
							PTR__->left1 = -1;
						else
						{
							PTR__->fade += PTR__->fadespd;
							PTR__->voldown = PTR__->fade >> FADE_SHIFT;
							SETTING_XYA(SetVolumeFM,         volFM);
							SETTING_XYA(SetVolumePSG,        volSSG);
							SETTING_XYA(SetVolumePCM,        volPCM);
#if 0
							SETTING_XYA(SetVolumeRhythmTotal,volRHYTHM);
#endif
						}
					}
					break;
				case 0xfd:	/* END */
				default:
					if (PTR__->fmtimersync && PTR__->left3)
					{
						PTR__->left1 = PTR__->left3;
						PTR__->left3 = 0;
						break;
					}
					if (PTR__->loopclock && PTR__->loopptr != NULL)
					{
#if EXLOOP
						int i;
						opn2.Reset();
						opn2.SetReg(0x29, 0x9f);
						PTR__->saveflagm[1] &= ~((1 << 0x08) + (1 << 0x0a));
						for (i = 0; i < 0x100; i++)
							if (PTR__->saveflagm[i >> 5] & (1 << (i & 31)))
								opn2.SetReg(i, PTR__->savem[i]);
						for (i = 0; i < 0x100; i++)
							if (PTR__->saveflags[i >> 5] & (1 << (i & 31)))
								opn2.SetReg(0x100 | i, PTR__->saves[i]);
						for (i = 0; i < 8; i++)
							if (!(PTR__->savekey[i] & 0x100))
								opn2.SetReg(0x28, PTR__->savekey[i]);
#endif
						p = PTR__->loopptr;
						PTR__->loopc++;
					}
					else
						PTR__->left1 = -1;
					break;
			}
		}

		if (PTR__->left1 == -1)
		{
			if (bufp != NULL && numsamp) ZeroMemory(bufp, numsamp*4);
			Stop(this__);
			PTR__->ptr = p;
			return;
		}

		while (PTR__->left2 > 60)
		{
			PTR__->left2 -= 60;
			PTR__->left1++;
		}

		if (bufp == NULL)
		{
			PTR__->left1 = 0;
			if (PTR__->clock * 100 >= numsamp * 6) break;
		}
		else
		{
			int psgvol, *mp;
			short *sp;
			size = min(PTR__->left1, numsamp);
			PTR__->left1   -= size;
			numsamp -= size;
			ZeroMemory(mixbuf, size*8);
			SN76496Update_16(0, snbuf, size);
			opn2.Mix(mixbuf, size);
			mp = mixbuf;
			sp = snbuf;
			psgvol = opn2.GetVolumeLinear();
			while (size--)
			{
				int samp;
				samp = *(mp++) + ((*(sp  ) * psgvol) >> 12);
				if (samp > 32767) samp = 32767;
				if (samp < -32768) samp = -32768;
				*(bufp++) = samp;
				samp = *(mp++) + ((*(sp++) * psgvol) >> 12);
				if (samp > 32767) samp = 32767;
				if (samp < -32768) samp = -32768;
				*(bufp++) = samp;
			}
		}
	}

	PTR__->ptr = p;

	if (PTR__->loopcount == 0 || PTR__->fade != 0) return;
	if (PTR__->loopc < PTR__->loopcount) return;
	if (PTR__->fadetime != 0)
	{
		PTR__->fade = 1;
		PTR__->fadespd = (((FADE_MAX * 100) << FADE_SHIFT) / 6) / PTR__->fadetime;
	}
	else
		Stop(this__);
	return;
}

static void Term(SEQUENCER *this__)
{
	if (PTR__->data != NULL) free(PTR__->data);
	free(PTR__);
}
static void SetPosition(SEQUENCER *this__, int time_in_ms)
{

	if (PTR__->clock * 100 > time_in_ms * 6) Play(this__);
	Mix(this__, NULL, time_in_ms);
	opn2.FlushPCM();
}

static int GetPosition(SEQUENCER *this__)
{
	return PTR__->clock * 100 / 6;
}
static int GetRate(SEQUENCER *this__)
{
	PTR__->frequency = setting.frequency;
	PTR__->frequency2 = (PTR__->frequency == SOUND_55K) ? SOUND_44K : PTR__->frequency;
	return PTR__->frequency2;
}
static int GetPriority(SEQUENCER *this__)
{
	return setting.priority;
}

static int GetLoopCount(SEQUENCER *this__)
{
	return PTR__->loopc;
}

static int IsPlaying(SEQUENCER *this__)
{
	return PTR__->isplaying;
}

static int GetLength(SEQUENCER *this__)
{
	int clocks = 0;
	unsigned char *p = PTR__->data;
	while (1)
	{
		switch (*p)
		{
			case 0x01:	/* OPNA(m) */
			case 0x02:	/* OPNA(s) */
				p += 3;
				break;
			case 0x03:	/* PSG */
				p += 2;
				break;
			case 0x00:	/* SYNC(1) */
				clocks += 1;
				p++;
				break;
			case 0xfd:	/* END */
			default:
				return clocks * 100 / 6;
		}
	}
}

#if SINGLE_BUILD
#define InitGEZSequencer InitSequencer
#define QuitGEZSequencer QuitSequencer
#define AboutGEZSequencer AboutSequencer
#define ConfigGEZSequencer ConfigSequencer
#define getGEZFileInfo getFileInfo
#define loadGEZFile loadFile
#define infoGEZBox infoBox
#define isOurGEZFile isOurFile
#endif

static void WritePrivateProfileInt(LPCSTR sect, LPCSTR key, UINT data, LPCSTR file)
{
	char temp[32], *p;
	temp[30] = '0';
	temp[31] = '\0';
	if (data == 0)
		p = &temp[30];
	else
		for (p = &temp[31]; data != 0; data /= 10)
		{
			*(--p) = '0' + (data % 10);
		}
	WritePrivateProfileString(sect, key, p, file);
}

static void GetSettingString(LPCSTR lpKey, LPCSTR lpDefault, LPSTR lpBuf, DWORD nSize)
{
	GetPrivateProfileString("GEZamp", lpKey, lpDefault, lpBuf, nSize, setting.cfgname);
}
static void SetSettingString(LPCSTR lpKey, LPCSTR lpStr)
{
	WritePrivateProfileString("GEZamp", lpKey, lpStr, setting.cfgname);
}
static int GetSettingInt(LPCSTR lpKey, int iDefault)
{
	char buf[128], dbuf[128];
	wsprintf(dbuf, "%d", iDefault);
	GetPrivateProfileString("GEZamp", lpKey, dbuf, buf, sizeof(buf), setting.cfgname);
	return atoi(buf);
}
static void SetSettingInt(LPCSTR lpKey, int iInt)
{
	char buf[128];
	wsprintf(buf, "%d", iInt);
	WritePrivateProfileString("GEZamp", lpKey, buf, setting.cfgname);
}

#include "cfgezamp.h"
#include "cfgezmix.h"

void InitGEZSequencer(HINSTANCE hDllInstance)
{
	char path[MAX_PATH];
	LPTSTR fnoff;

	setting.hDllInstance = hDllInstance;

	GetModuleFileName(hDllInstance, path, MAX_PATH);
	GetFullPathName(path, MAX_PATH, setting.cfgname, &fnoff);
	setting.cfgname[(fnoff - setting.cfgname)] = '\0';
	lstrcat(setting.cfgname, "plugin.ini");

	setting.fmtimersync = 0;

	GEZAMPLoadSetting();
	GEZMIXLoadSetting();
}

void QuitGEZSequencer(HINSTANCE hDllInstance)
{
}

void AboutGEZSequencer(HWND hwndParent)
{
	MessageBox(hwndParent, "M88 - PC-88 Emulator | FM Sound Generator - Core Unit\n\tCopyright (C) cisc 1998, 2000.\n\nZLIB 1.1.3\n\t(C) 1995-1998 Jean-loup Gailly and Mark Adler\n\nGEZamp\n\t(C) 2000 Mamiya\n\t\tmailto:mamiya@proc.org.tohoku.ac.jp\n\t\thttp://www.proc.org.tohoku.ac.jp/befis/", PLUGIN_NAME, MB_OK);
}

void ConfigGEZSequencer(HWND hwndParent)
{
	PROPSHEETPAGE psp[2];
	PROPSHEETHEADER psh;

	ZeroMemory(psp, sizeof(psp));
	psp[0].dwSize = sizeof(PROPSHEETPAGE);
	psp[0].dwFlags = PSP_DEFAULT;
	psp[0].hInstance = setting.hDllInstance;
	psp[0].pszTemplate = MAKEINTRESOURCE(IDD_PSP_GEZAMP);
	psp[0].pszIcon = NULL;
	psp[0].pszTitle = NULL;
	psp[0].pfnDlgProc = GEZAMPSettingProc;
	psp[0].lParam = 0;
	psp[0].pfnCallback = NULL;
	psp[0].pcRefParent = NULL;

	psp[1].dwSize = sizeof(PROPSHEETPAGE);
	psp[1].dwFlags = PSP_DEFAULT;
	psp[1].hInstance = setting.hDllInstance;
	psp[1].pszTemplate = MAKEINTRESOURCE(IDD_PSP_GEZMIX);
	psp[1].pszIcon = NULL;
	psp[1].pszTitle = NULL;
	psp[1].pfnDlgProc = GEZMIXSettingProc;
	psp[1].lParam = 0;
	psp[1].pfnCallback = NULL;
	psp[1].pcRefParent = NULL;

	ZeroMemory(&psh, sizeof(psh));
	psh.dwSize = sizeof(PROPSHEETHEADER);
	psh.dwFlags = PSH_NOAPPLYNOW | PSH_PROPSHEETPAGE;
	psh.hwndParent = hwndParent;
	psh.hInstance = setting.hDllInstance;
	psh.pszIcon = NULL;
	psh.pszCaption = (LPSTR) "GEZamp setting";
	psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
	psh.nStartPage = 0;
	psh.ppsp = (LPCPROPSHEETPAGE) &psp;
	psh.pfnCallback = NULL;
	PropertySheet(&psh);
}

static int GetDwordLE(unsigned char *p)
{
	int ret;
	ret  = ((int)(unsigned char)p[3]) << 0x18;
	ret |= ((int)(unsigned char)p[2]) << 0x10;
	ret |= ((int)(unsigned char)p[1]) << 0x08;
	ret |= ((int)(unsigned char)p[0]) << 0x00;
	return ret;
}

SEQUENCER *loadGEZFile(char *fn)
{
	FILE *fp = NULL;
	GEZSEQ *p = NULL;
	while (1)
	{
		int size;
		p = (GEZSEQ *) malloc(sizeof(GEZSEQ));
		if (p == NULL) break;;

		p->seq.work = p;
		p->title[0] = '\0';
		p->loopclock = 0;
#if USE_ZLIB
		p->data = (unsigned char *)GEZ_extract(fn, &size, &p->loopclock, p->title);
#else
		p->data = NULL;
#endif
		GYI_load(fn, &p->loopclock, p->title);
		if (p->data == NULL)
		{
			size_t flen;
			fp = fopen(fn, "rb");
			if (fp == NULL) break;
			fseek(fp, 0, SEEK_END);
			size = flen= ftell(fp);
			fseek(fp, 0, SEEK_SET);
			p->data = (unsigned char *)malloc(flen + 1);
			if (p->data == NULL) break;
			if (flen != fread(p->data, 1, flen, fp)) break;
			fclose(fp); fp = NULL;
		}

		if (GetDwordLE(p->data + GYM_OFS_MAGIC) == GYM_MAGIC)
		{
			if (*p->title == '\0')	/* GYID */
			{
				int len;
				char *dp = p->title;
				unsigned char *sp = p->data + GYM_OFS_GAMETITLE;
				for (len = 0; *sp && len < 32; len++) *(dp++) = *(sp++);
				sp = p->data + GYM_OFS_SONGTITLE;
				if (*sp)
				{
					if (dp != p->title) *(dp++) = ' ';
					*(dp++) = '[';
					*(dp++) = ' ';
					for (len = 0; *sp && len < 32; len++) *(dp++) = *(sp++);
					*(dp++) = ' ';
					*(dp++) = ']';
				}
				*dp = '\0';
			}
			if (p->loopclock == 0)	/* GYID */
				p->loopclock = GetDwordLE(p->data + GYM_OFS_LOOPTIME);
			if (GetDwordLE(p->data + GYM_OFS_COMPSIZE))
			{
#if USE_ZLIB
				uLongf dessize;
				unsigned char *des;
				des = (unsigned char *)malloc(GetDwordLE(p->data + GYM_OFS_COMPSIZE) + 1);
				if (des == NULL) break;
				if (Z_OK != uncompress(des, &dessize, p->data + GYM_OFS_DATASTART, size - GYM_OFS_DATASTART)) {
					free(des);
					break;
				}
				free(p->data);
				size = dessize;
				p->data = des;
#else
				break;	/* NOT SUPPORT */
#endif
			}
			else
			{
				/* YMAMPwb_폜 */
				size -= GYM_OFS_DATASTART;
				MoveMemory(p->data, p->data + GYM_OFS_DATASTART, size);
			}
		}
		p->data[size] = '\xff';	/* ENDMARK */

		p->clock = 0;
		p->left1 = 0;
		p->ptr = NULL;
		p->loopc = 0;
		p->isplaying = 0;

		p->seq.Play = Play;
		p->seq.Stop = Stop;
		p->seq.Mix = Mix;
		p->seq.Term = Term;
		p->seq.SetPosition = SetPosition;
		p->seq.GetPosition = GetPosition;
		p->seq.GetRate = GetRate;
		p->seq.GetPriority = GetPriority;
		p->seq.GetLoopCount = GetLoopCount;
		p->seq.IsPlaying = IsPlaying;
		return &p->seq;
	}
	if (fp != NULL) fclose(fp);
	if (p != NULL) Term(&p->seq);
	return NULL;
}

int getGEZFileInfo(char *fn, char *title, int *length_in_ms)
{
	SEQUENCER *this__;
	this__ = loadGEZFile(fn);
	if (this__ == NULL) return FALSE;
	if (title != NULL)
	{
		lstrcpy(title, PTR__->title);
		if (*title == '\0')
		{
			char *p, *p2;
			for (p2 = p = fn;1;)
			{
				char *prev;
				if (*p == '/' || *p == '\\' || *p == ':') p2 = p + 1;
				p = CharNext(prev = p);
				if (p == prev) break;
			}
			lstrcpy(title, "[ ");
			lstrcat(title, p2);
			lstrcat(title, " ]");
		}
	}
	if (length_in_ms != NULL)
		*length_in_ms = GetLength(this__);
	Term(this__);
	return true;
}

int infoGEZBox(char *fn, HWND hwndParent)
{
	return 0;
}

int isOurGEZFile(char *fn)
{
	if (fn)
	{
		int l = lstrlen(fn) - 8;
		if (l >= 0 && lstrcmpi(fn + l, ".GYM.ZIP") == 0) return 1;
	}
	return 0;
}
